/*
 * Oleg Moskalenko, 2013
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/fcntl.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <inttypes.h>
#include <errno.h>

#define SERVER_ADDRESS ("127.0.0.1")
#define SERVER_PORT (3478)

#define CLIENT_ADDRESS ("127.0.0.1")
#define CLIENT_PORT (3479)

#define PACKET1 ("p1")
#define PACKET2 ("p2")
#define PACKET3 ("p3")

typedef union {
  struct sockaddr_storage ss;
  struct sockaddr_in s4;
} ioa_addr;

static int socket_set_reusable(int fd) {
	if (fd < 0)
		return -1;
	else {

#ifdef SO_REUSEPORT
		{
			const int on = 1;
			int ret = setsockopt(fd, SOL_SOCKET, SO_REUSEPORT,
					(const void*) &on, (socklen_t) sizeof(on));
			if (ret < 0)
				perror(__FUNCTION__);
		}
#endif

#ifdef SO_REUSEADDR
		{
			const int on = 1;
			int ret = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
					(const void*) &on, (socklen_t) sizeof(on));
			if (ret < 0)
				perror(__FUNCTION__);
		}
#endif

		return 0;
	}
}

static int addr_connect(int fd, const ioa_addr* addr)
{
	if (!addr || fd < 0)
		return -1;
	else {
		int err = 0;
		do {
			if (addr->ss.ss_family == AF_INET) {
				err = connect(fd, (const struct sockaddr *) addr, sizeof(struct sockaddr_in));
			} else {
				return -1;
			}
		} while (err < 0 && errno == EINTR);

		if (err < 0 && errno != EINPROGRESS) {
			perror("Connect");
			fprintf(stderr,"fatal error: cannot connect socket.\n");
			exit(-1);
		}

		return err;
	}
}

static int addr_bind(int fd, const ioa_addr* addr)
{
	if (!addr || fd < 0)
		return -1;
	else {
		int ret = -1;

		socket_set_reusable(fd);

		if (addr->ss.ss_family == AF_INET) {
			do {
				ret = bind(fd, (const struct sockaddr *) addr, sizeof(struct sockaddr_in));
			} while (ret < 0 && errno == EINTR);
		} else {
			return -1;
		}

		if(ret<0) {
			perror("bind");
			fprintf(stderr,"fatal error: cannot bind socket.\n");
			exit(-1);
		}

		return ret;
	}
}

static int udp_send(int fd, const ioa_addr* dest_addr, const char* buffer, int len)
{
	int rc = 0;
	if(dest_addr) {
		socklen_t slen = sizeof(struct sockaddr_in);
		do {
			rc = sendto(fd, buffer, len, 0, (const struct sockaddr*) dest_addr, slen);
		} while (rc < 0 && (errno == EINTR));
	} else {
		do {
			rc = send(fd, buffer, len, 0);
		} while (rc < 0 && (errno == EINTR));
	}
	if(rc<0 && ((errno == ENOBUFS) || (errno == EAGAIN))) {
		//Lost packet
		rc = len;
	}

	if(rc<0)
		perror("send");

	return rc;
}

static int udp_recvfrom(int fd, ioa_addr* addr, char* buffer, int buf_size)
{
	if (fd < 0 || !addr || !buffer)
		return -1;

	int len = -1;
	socklen_t slen = sizeof(struct sockaddr_in);

	do {
	  len = recvfrom(fd, buffer, buf_size, 0, (struct sockaddr*) addr, (socklen_t*) &slen);
	} while (len < 0 && (errno == EINTR));

	return len;
}

static void init_addr(const char* saddr, int port, ioa_addr *addr)
{
	inet_pton(AF_INET, saddr, &(addr->s4.sin_addr));
    addr->s4.sin_family = AF_INET;
#if defined(__BSD_VISIBLE)
    addr->s4.sin_len = sizeof(struct sockaddr_in);
#endif
    addr->s4.sin_port = htons(port);
}

static void init_addrs(ioa_addr *server_addr, ioa_addr *client_addr)
{
	init_addr(SERVER_ADDRESS,SERVER_PORT,server_addr);
	init_addr(CLIENT_ADDRESS,CLIENT_PORT,client_addr);
}

static int open_socket() {
	int fd = socket(AF_INET, SOCK_DGRAM, 0);
	socket_set_reusable(fd);
	return fd;
}

static int addr_eq(const ioa_addr* a1, const ioa_addr *a2) {
	if (!a1)
		return (!a2);

	if (a1->ss.ss_family == a2->ss.ss_family) {
		if (a1->ss.ss_family == AF_INET) {
			if ((int) a1->s4.sin_addr.s_addr == (int) a2->s4.sin_addr.s_addr
					&& a1->s4.sin_port == a2->s4.sin_port) {
				return 1;
			}
		}
	}

	return 0;
}

